#!/bin/bash
# k8s_backup.sh

set -euo pipefail

# Check for required environment variables
postgres_vars=("POSTGRES_HOST" "POSTGRES_PORT" "POSTGRES_USER" "POSTGRES_PASSWORD")
minio_vars=("AWS_CONNECTION_STRING" "AWS_ACCESS_KEY" "AWS_SECRET_KEY")

# Function to check if variables are set and not empty
check_required_variables() {
  local id="$1"
  shift
  local target_vars=("$@")
  local unset_vars=()
  for var in "${target_vars[@]}"; do
    if [[ -z "${!var:-}" ]]; then
      unset_vars+=("$var")
    fi
  done

  if [[ ${#unset_vars[@]} -gt 0 ]]; then
    echo "Please set the following environment variables in '$id' : ${unset_vars[*]}"
    exit 1
  fi
}
# Check for required dependencies
command_check() {
  local required_command="$1"
  command -v "$required_command" >/dev/null 2>&1 || {
    echo "'$required_command' command not found. Make sure it's installed and in the PATH."
    exit 1
  }
}

# Validate input
validate_input() {
  if [ $# -ne 1 ]; then
    echo "Usage: $0 <namespace>"
    exit 1
  fi
}

# Validate the existence of the specified namespace
validate_namespace() {
    local namespace="$1"

    # Check if the provided namespace exists
    if ! kubectl get namespace "$namespace" &> /dev/null; then
        echo "Namespace '$namespace' not found."
        exit 1
    fi
}

# Compatibility for Windows
pwd_compatible() {
  if [[ "$(expr substr $(uname -s) 1 5)" == "MINGW" ]]; then
    work_dir=$(pwd -W)
  else
    work_dir=$(pwd)
  fi
}

# Function to stop deployments
stop_deployments() {
  local s_deployments=("$@")

  for deployment in "${s_deployments[@]}"; do
    replicas["$deployment"]=$(kubectl get deployment "$deployment" -n "$namespace" -o=jsonpath='{.spec.replicas}')
    kubectl -n "$namespace" scale deployment "$deployment" --replicas 0 > /dev/null 2>&1
  done
  for deployment in "${s_deployments[@]}"; do
    while ! [ -z "$(kubectl get deployment $deployment -n $namespace -o=jsonpath='{.status.readyReplicas}')" ]; do
      sleep 1
    done
  done
}

# Function to start deployments
start_deployments() {
  local s_deployments=("$@")

  for deployment in "${s_deployments[@]}"; do
    kubectl scale deployment "$deployment" -n "$namespace" --replicas "${replicas["$deployment"]}" > /dev/null 2>&1
  done

  for deployment in "${s_deployments[@]}"; do
    local counter=0
    local timeout=180
    while [[ "$(kubectl get deployment "$deployment" -n "$namespace" -o=jsonpath='{.status.readyReplicas}')" != "${replicas["$deployment"]}" ]]; do
      sleep 1
      ((counter+=1))
      if [[ "$counter" == "$timeout" ]]; then
        echo "Deployment '$deployment' is taking longer than ${timeout}s to become ready."
        break
      fi
    done
  done
}

# Function to find the statefulset with matching image
find_statefulset_with_image() {
    local image_partial="$1"
    local statefulset=$(kubectl get statefulsets -n $namespace -o=jsonpath='{range .items[*]}{@.metadata.name}{" "}{@.spec.template.spec.containers[*].image}{"\n"}{end}' | grep "$image_partial" | awk 'NR==1{print $1}')

    if [[ -z $statefulset ]]; then
        exit 1
    else
        echo "${statefulset}-0"
    fi
}

# Function to run k8s job and track its completion
run_job() {
  local job_name="$1"
  local job_path="$2"
  local dest="$3"
  local arch="$4"
  envsubst < "${job_path}/job.yaml" > ./tmp.yaml
  kubectl apply -f "${job_path}/configmap.yaml" -n "$namespace" > /dev/null 2>&1
  kubectl apply -f ./tmp.yaml -n "$namespace" > /dev/null 2>&1

  # Wait for the job to complete or fail using the loop with sleep
  while true; do
    # Get the name of the pod associated with the job
    local pod_name=$(kubectl get pods -n "$namespace" -l job-name="$job_name" -o jsonpath='{.items[0].metadata.name}')
    local succeeded_status=$(kubectl logs "$pod_name" -n "$namespace" | grep -c "Script finished.")
    local failed_status=$(kubectl get job "$job_name" -n "$namespace" -o=jsonpath='{.status.failed}')
    
    if [ "$succeeded_status" = "1" ]; then
      echo "Job '$job_name' successful."
      copy_from_pod "$pod_name" "$arch" "$dest"
      kubectl logs "$pod_name" -n "$namespace"
      kubectl delete -f ./tmp.yaml -n "$namespace" 2>&1
      kubectl delete -f "${job_path}/configmap.yaml" -n "$namespace" > /dev/null 2>&1
      rm -f ./tmp.yaml
      break
    elif [ "$failed_status" = "1" ]; then
      echo "Job '$job_name' failed. Displaying logs:"
      # Display the logs of the failed pod
      kubectl logs "$pod_name" -n "$namespace"
      kubectl delete -f ./tmp.yaml -n "$namespace" 2>&1
      kubectl delete -f "${job_path}/configmap.yaml" -n "$namespace" > /dev/null 2>&1
      rm -f ./tmp.yaml
      exit 1
    fi
    sleep 1
  done
}

# Function to copy data from pod
copy_from_pod() {
  local pod_name="$1"
  local arch="$2"
  local dest="$3"
  kubectl cp "${pod_name}:/tmp/${arch}" "${dest}/${arch}" -n "$namespace" --retries 999 1>/dev/null 2>&1
  kubectl exec "$pod_name" -n "$namespace" -- rm -rf tmp/"$arch"
}

backup_postgres() {
  local target_arch="postgres_${arch_name}"
  # Run backup job
  run_job "postgres-backup-job" "./jobs/postgres/backup" "$backup_path" "$target_arch" 
  echo "Backup for postgres created in $target_arch."
}

backup_minio() {
  local target_arch="minio_${arch_name}"
  # Run backup job
  run_job "minio-backup-job" "./jobs/minio/backup" "$backup_path" "$target_arch" 
  echo "Backup for minio created in $target_arch."
}

# Function to backup services
backup_services() {
  local redis_arch="${backup_path}/redis_${arch_name}"
  local influx_arch="${backup_path}/influxdb_${arch_name}"

  local redis_pod=$(find_statefulset_with_image "redis")
  local influx_pod=$(find_statefulset_with_image "influxdb")

  # Copy backup from redis pod
  kubectl exec "$redis_pod" -n "$namespace" -- tar -C /data -cf /tmp/redis.tar .
  kubectl cp "$redis_pod:/tmp/redis.tar" "$redis_arch" -n "$namespace" --retries 999
  kubectl exec "$redis_pod" -n "$namespace" -- rm -f /tmp/redis.tar

  # Copy backup from influxdb pod
  kubectl exec "$influx_pod" -n "$namespace" -- tar -C /var/lib/influxdb -cf /tmp/influxdb.tar .
  kubectl cp "$influx_pod:/tmp/influxdb.tar" "$influx_arch" -n "$namespace" --retries 999
  kubectl exec "$redis_pod" -n "$namespace" -- rm -f /tmp/influxdb.tar

  echo "Data copied from 'auth-cache' and 'influxdb'."
}

main() {
  # Main script logic
  echo "Starting backup process for namespace '$namespace'..."

  # Variables for backup
  backup_date=$(date +"%d_%m_%Y")
  arch_name="backup.tar"
  pwd_compatible
  backup_path="${work_dir}/backups/${backup_date}"
  mkdir -p "$backup_path" || true

  # Stop deployments
  echo "Stopping TestIT..."
  declare -A replicas
  export replicas
  local testit=("frontend" "webapi" "background-service" "rabbitmqconsumer" "auth" "avatars-api" "ldapwebapi" "license-service" "globalsearch-service")
  stop_deployments "${testit[@]}"

  # Backup services
  echo "Backing up services..."
  backup_services

  # Backup databases
  echo "Backing up databases..."
  backup_postgres

  # Backup minio
  echo "Backing up buckets..."
  backup_minio

  # Start deployments
  echo "Starting TestIT..."
  start_deployments "${testit[@]}"

  echo "Backup process completed successfully!"
  exit 0
}

check_required_variables "postgres_vars" "${postgres_vars[@]}"
check_required_variables "minio_vars" "${minio_vars[@]}"
command_check "kubectl"
command_check "tar"
validate_input "$@"
namespace="$1"
validate_namespace "$namespace"

main "$@"
exit 0