#!/bin/bash
# move_to_k8s.sh

set -euo pipefail

# Array of required environment variables
minio_vars=("AWS_CONNECTION_STRING" "TESTIT_BUCKET" "AVATARS_BUCKET")
postgres_vars=("POSTGRES_HOST" "POSTGRES_PORT")

# Function to check if variables are set and not empty
check_required_variables() {
  local id="$1"
  shift
  local target_vars=("$@")
  local unset_vars=()
  for var in "${target_vars[@]}"; do
    if [[ -z "${!var:-}" ]]; then
      unset_vars+=("$var")
    fi
  done

  if [[ ${#unset_vars[@]} -gt 0 ]]; then
    echo "Please set the following environment variables in '$id' : ${unset_vars[*]}"
    exit 1
  fi
}
# Check for required dependencies
command_check() {
  local required_command="$1"
  command -v "$required_command" >/dev/null 2>&1 || {
    echo "'$required_command' command not found. Make sure it's installed and in the PATH."
    exit 1
  }
}

# Validate input
validate_input() {
  if [ $# -ne 2 ]; then
    echo "Usage: $0 <namespace> </path/to/archive.tar>"
    exit 1
  fi
}

# Validate the existence of the specified namespace
validate_namespace() {
    local namespace="$1"

    # Check if the provided namespace exists
    if ! kubectl get namespace "$namespace" &> /dev/null; then
        echo "Namespace '$namespace' not found."
        exit 1
    fi
}

# Compatibility for Windows
pwd_compatible(){
  if [[ "$(expr substr $(uname -s) 1 5)" == "MINGW" ]]; then
    work_dir=$(pwd -W)
  else
    work_dir=$(pwd)
  fi
}

# Function to stop deployments
stop_deployments() {
  local s_deployments=("$@")

  for deployment in "${s_deployments[@]}"; do
    replicas["$deployment"]=$(kubectl get deployment "$deployment" -n "$namespace" -o=jsonpath='{.spec.replicas}')
    kubectl -n "$namespace" scale deployment "$deployment" --replicas 0 > /dev/null 2>&1
  done

  for deployment in "${s_deployments[@]}"; do
    while ! [ -z "$(kubectl get deployment $deployment -n $namespace -o=jsonpath='{.status.readyReplicas}')" ]; do
      sleep 1
    done
  done
}

# Function to start deployments
start_deployments() {
  local s_deployments=("$@")

  for deployment in "${s_deployments[@]}"; do
    kubectl scale deployment "$deployment" -n "$namespace" --replicas "${replicas["$deployment"]}" > /dev/null 2>&1
  done

  for deployment in "${s_deployments[@]}"; do
    local counter=0
    local timeout=180
    if [ "${replicas["$deployment"]}" -gt 0 ]; then
     while [[ "$(kubectl get deployment "$deployment" -n "$namespace" -o=jsonpath='{.status.readyReplicas}')" != "${replicas["$deployment"]}" ]]; do
       sleep 1
       ((counter+=1))
       if [[ "$counter" == "$timeout" ]]; then
         echo "Deployment '$deployment' is taking longer than ${timeout}s to become ready."
         break
       fi
     done
    fi
   done
}

# Function to check if the file exists
check_file_existence() {
  local filename="$1"
  if ! [ -f "$filename" ]; then
    echo "The file '$filename' is missing."
    echo "Please check your working directory and make sure the file is present."
  fi
}

# Function to check if a directory exists
check_directory_existence() {
  local directory_name="$1"
  if ! [ -d "$directory_name" ]; then
    echo "The directory '$directory_name' is missing."
    echo "Please check your working directory and make sure the directory is present."
    exit 1
  fi
}

# Function to find the statefulset with matching image
find_statefulset_with_image() {
    local image_partial="$1"
    local statefulset=$(kubectl get statefulsets -n $namespace -o=jsonpath='{range .items[*]}{@.metadata.name}{" "}{@.spec.template.spec.containers[*].image}{"\n"}{end}' | grep "$image_partial" | awk 'NR==1{print $1}')

    if [[ -z $statefulset ]]; then
        exit 1
    else
        echo "${statefulset}-0"
    fi
}

# Function to get an environment variable from a Kubernetes pod
get_env_variable_from_pod() {
  local statefulset="$1"
  local variable_name="$2"
  local pod_name=$(find_statefulset_with_image "$statefulset")
  if [ -z "$pod_name" ]; then
    exit 1
  fi
  # Use kubectl exec to run the printenv command inside the pod and fetch the environment variable
  local env_value=$(kubectl exec -n "$namespace" "$pod_name" -- printenv "$variable_name" 2>/dev/null)

  # Output the value of the environment variable
  echo "$env_value"
}

# Function to copy data to pod
copy_to_pod() {
  local pod="$1"
  local file="$2"
  local dest="$3"
 
  kubectl cp  -n "$namespace" "${file}" "${pod}:${dest}" --retries 999 --no-preserve > /dev/null 2>&1
}

# Function to run k8s job and track its completion
run_job() {
  local job_name="$1"
  local job_path="$2"
  local source="$3"
  local arch="$4"

  # Create archive
  tar -C "$source" -cf "${arch}" .

  envsubst < "${job_path}/job.yaml" > ./tmp.yaml
  kubectl apply -f "${job_path}/configmap.yaml" -n "$namespace" > /dev/null 2>&1
  kubectl apply -f ./tmp.yaml -n "$namespace" > /dev/null 2>&1
  # Get the name of the pod associated with the job
  local pod_name=""
  echo "Watging for job '$job_name' to start..."
  local status=0
  while [ $status -lt 1 ]; do
    pod_name=$(kubectl get pods -n "$namespace" -l job-name="$job_name" -o jsonpath='{.items[0].metadata.name}')
    if [ "$(kubectl get pods -n "$namespace" "$pod_name" -o jsonpath='{.status.containerStatuses[0].ready}')" == "true" ]; then
      status=$(kubectl logs "$pod_name" -n "$namespace" | grep -c "Waiting")
    fi
    sleep 1
  done
  # Copy backup files to pod
  copy_to_pod "$pod_name" "$arch" "/tmp/archive.tar"
  echo "Init" > "${buffer_dir}/start.txt"
  copy_to_pod "$pod_name" "${buffer_dir}/start.txt" "/tmp/start.txt"
  # Wait for the job to complete or fail using the loop with sleep
  while true; do
    local succeeded_status=$(kubectl logs "$pod_name" -n "$namespace" | grep -c "Script finished.")
    local failed_status=$(kubectl get job "$job_name" -n "$namespace" -o=jsonpath='{.status.failed}')
    
    if [ "$succeeded_status" = "1" ]; then
      echo "Job '$job_name' successful."
      # Display the logs of the job
      kubectl logs "$pod_name" -n "$namespace" | grep -v "Waiting"
      kubectl delete -f ./tmp.yaml -n "$namespace" > /dev/null 2>&1
      kubectl delete -f "${job_path}/configmap.yaml" -n "$namespace" > /dev/null 2>&1
      rm -f ./tmp.yaml
      break
    elif [ "$failed_status" = "1" ]; then
      echo "Job '$job_name' failed. Displaying logs:"
      # Display the logs of the job
      kubectl logs "$pod_name" -n "$namespace"
      kubectl delete -f ./tmp.yaml -n "$namespace" > /dev/null 2>&1
      kubectl delete -f "${job_path}/configmap.yaml" -n "$namespace" > /dev/null 2>&1
      rm -f ./tmp.yaml
      exit 1
    fi
    sleep 1
  done
}

# Function to upload Redis and InfluxDB data
upload_redis_influx() {
  echo "Uploading Redis + InfluxDB..."
  local influxdb_arch="${buffer_dir}/influxdb.tar"
  local redis_dir="${buffer_dir}/backup/redis"
  local influxdb_dir="${buffer_dir}/backup/influxdb"
  check_directory_existence "$redis_dir"
  check_directory_existence "$influxdb_dir"

  local redis_pod=$(find_statefulset_with_image "redis")
  local influx_pod=$(find_statefulset_with_image "influxdb")

  # Copy backup files for redis
  for file in $(ls "$redis_dir"); do
    local full_path="${redis_dir}/${file}"
    kubectl cp "$full_path" "$redis_pod:/data/${file}" -n "$namespace" --retries 999 --no-preserve
  done

  tar -C "$influxdb_dir" -cf "$influxdb_arch" .

  # Copy backup archives to influxdb pod and unpack
  kubectl cp "$influxdb_arch" "${influx_pod}:tmp/influxdb.tar" -n "$namespace" --retries 999 --no-preserve
  kubectl exec "$influx_pod" -n "$namespace" -- tar -C var/lib/influxdb -xf tmp/influxdb.tar
  kubectl exec "$influx_pod" -n "$namespace" -- rm -f tmp/influxdb.tar

  echo "Data restored to 'auth-cache' and 'influxdb'."
}

# Function to upload to minio server form docker backups
upload_minio() {
  echo "Uploading buckets..."
  local minio_dir="$buffer_dir/backup/minio"
  check_directory_existence "${minio_dir}/testit"
  check_directory_existence "${minio_dir}/testit_configs"
  check_directory_existence "${minio_dir}/avatars"
  
  run_job "mc-upload-job" "./jobs/minio/docker-to-k8s" "$minio_dir" "${buffer_dir}/minio.tar"
}

# Function to restore databases from docker backups
upload_postgres() {
  echo "Uploading databases..."
  local postgres_dir="$buffer_dir/backup/postgres"
  check_directory_existence "$postgres_dir"
  local databases=("testitdb" "authdb" "avatarsdb" "backgrounddb" "licensedb" "globalsearchdb")
  for db in "${databases[@]}"; do
    check_file_existence "$postgres_dir/${db}_backup.sql"
  done

  run_job "postgres-export-job" "./jobs/postgres/docker-to-k8s" "$postgres_dir" "${buffer_dir}/postgres.tar"
}

main () {
  # Main script logic
  echo "Starting restore process for namespace '$namespace' from docker backup..."

  pwd_compatible
  export buffer_dir="${work_dir}/buffer"
  mkdir -p "$buffer_dir"
  local archive="$2"
  tar -xf "$archive" -C "$buffer_dir" 

  # Save unpacked trusted certificates
  rm -rf "$work_dir/CERTS/"
  mkdir -p "$work_dir/CERTS/"
  mv "${buffer_dir}/backup/certs/ca-certificates/" "$work_dir/CERTS/" 
  echo "Trusted certificates saved at '$work_dir/CERTS/'."

  # Stop deployments
  declare -A replicas
  export replicas
  local testit=("frontend" "webapi" "auth" "avatars-api" "background-service" "ldapwebapi" "license-service" "rabbitmqconsumer" "globalsearch-service" )
  echo "Stopping Test IT..."
  stop_deployments "${testit[@]}"

  # Get minio creds
  export AWS_ACCESS_KEY=$(get_env_variable_from_pod "minio" "MINIO_ACCESS_KEY")
  export AWS_SECRET_KEY=$(get_env_variable_from_pod "minio" "MINIO_SECRET_KEY")
  upload_minio 

  # Get posrtges creds
  export POSTGRES_USER=$(get_env_variable_from_pod "postgres" "POSTGRES_USER")
  export POSTGRES_PASSWORD=$(get_env_variable_from_pod "postgres" "POSTGRES_PASSWORD")
  upload_postgres

  upload_redis_influx

  rm -rf "$buffer_dir"

  # Start deployments
  echo "Starting Test IT..."
  start_deployments "${testit[@]}"

  echo "Restore process completed successfully! Your Test IT is running on Kubernetes now!"
  echo "If you have SSL certificates (trusted-certificates-volume) or any other configurations to apply, you can reference the Test IT docs for guides."
  exit 0
}

command_check "kubectl"
command_check "tar"
check_required_variables "minio_vars" "${minio_vars[@]}"
check_required_variables "postgres_vars" "${postgres_vars[@]}"

validate_input "$@"
namespace="$1"
validate_namespace "$namespace"
main "$@"
